@@ -1,7 +1,7 @@
 /* trap.c -- Not the trap command, but useful functions for manipulating
    those objects.  The trap command is in builtins/trap.def. */
 
-/* Copyright (C) 1987-2020 Free Software Foundation, Inc.
+/* Copyright (C) 1987-2021 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -83,11 +83,11 @@ static void free_trap_string (int);
 static void reset_signal (int);
 static void restore_signal (int);
 static void reset_or_restore_signal_handlers (sh_resetsig_func_t *);
+static void reinit_trap (int);
 
 static void trap_if_untrapped (int, char *);
 
 /* Variables used here but defined in other files. */
-extern procenv_t alrmbuf;
 
 extern volatile int from_return_trap;
 extern int waiting_for_child;
@@ -297,11 +297,13 @@ void
 run_pending_traps ()
 {
   register int sig;
-  int old_exit_value, x;
-  int old_running;
+  int x;
+  volatile int old_exit_value, old_running;
   WORD_LIST *save_subst_varlist;
   HASH_TABLE *save_tempenv;
   sh_parser_state_t pstate;
+  volatile int save_return_catch_flag, function_code;
+  procenv_t save_return_catch;
 #if defined (ARRAY_VARS)
   ARRAY *ps;
 #endif
@@ -311,9 +313,7 @@ run_pending_traps ()
 
   if (running_trap > 0)
     {
-#if defined (DEBUG)
-      internal_warning ("run_pending_traps: recursive invocation while running trap for signal %d", running_trap-1);
-#endif
+      internal_debug ("run_pending_traps: recursive invocation while running trap for signal %d", running_trap-1);
 #if defined (SIGWINCH)
       if (running_trap == SIGWINCH+1 && pending_traps[SIGWINCH])
 	return;			/* no recursive SIGWINCH trap invocations */
@@ -343,9 +343,7 @@ run_pending_traps ()
 	 while (pending_traps[sig]--) instead of the if statement. */
       if (pending_traps[sig])
 	{
-	  if (running_trap == sig+1)
-	    /*continue*/;
-
+	  /* XXX - set last_command_exit_value = trap_saved_exit_value here? */
 	  running_trap = sig + 1;
 
 	  if (sig == SIGINT)
@@ -421,7 +419,6 @@ run_pending_traps ()
 	    }
 	  else
 	    {
-	      /* XXX - should we use save_parser_state/restore_parser_state? */
 	      save_parser_state (&pstate);
 	      save_subst_varlist = subst_assign_varlist;
 	      subst_assign_varlist = 0;
@@ -434,7 +431,23 @@ run_pending_traps ()
 	      /* XXX - set pending_traps[sig] = 0 here? */
 	      pending_traps[sig] = 0;
 	      evalnest++;
-	      evalstring (savestring (trap_list[sig]), "trap", SEVAL_NONINT|SEVAL_NOHIST|SEVAL_RESETLINE);
+
+	      function_code = 0;
+	      save_return_catch_flag = return_catch_flag;
+	      if (return_catch_flag)
+		{
+		  COPY_PROCENV (return_catch, save_return_catch);
+		  function_code = setjmp_nosigs (return_catch);
+		}
+
+	      if (function_code == 0)
+		x = parse_and_execute (savestring (trap_list[sig]), "trap", SEVAL_NONINT|SEVAL_NOHIST|SEVAL_RESETLINE);
+	      else
+		{
+		  parse_and_execute_cleanup (sig + 1);	/* XXX - could use -1 */
+		  x = return_catch_value;
+		}
+
 	      evalnest--;
 #if defined (JOB_CONTROL)
 	      restore_pipeline (1);
@@ -443,6 +456,19 @@ run_pending_traps ()
 	      subst_assign_varlist = save_subst_varlist;
 	      restore_parser_state (&pstate);
 	      temporary_env = save_tempenv;
+
+	      if (save_return_catch_flag)
+		{
+		  return_catch_flag = save_return_catch_flag;
+		  return_catch_value = x;
+		  COPY_PROCENV (save_return_catch, return_catch);
+		  if (function_code)
+		    {
+		      running_trap = old_running;		/* XXX */
+		      /* caller will set last_command_exit_value */
+		      sh_longjmp (return_catch, 1);
+		    }
+		}
 	    }
 
 	  pending_traps[sig] = 0;	/* XXX - move before evalstring? */
@@ -475,9 +501,7 @@ trap_handler (sig)
 
   if ((sigmodes[sig] & SIG_TRAPPED) == 0)
     {
-#if defined (DEBUG)
-      internal_warning ("trap_handler: signal %d: signal not trapped", sig);
-#endif
+      internal_debug ("trap_handler: signal %d: signal not trapped", sig);
       SIGRETURN (0);
     }
 
@@ -570,7 +594,7 @@ any_signals_trapped ()
   register int i;
 
   for (i = 1; i < NSIG; i++)
-    if (sigmodes[i] & SIG_TRAPPED)
+    if ((sigmodes[i] & SIG_TRAPPED) && (sigmodes[i] & SIG_IGNORED) == 0)
       return i;
   return -1;
 }
@@ -587,7 +611,8 @@ clear_pending_traps ()
 void
 check_signals ()
 {
-  CHECK_ALRM;		/* set by the read builtin */
+  /* Add any other shell timeouts here */
+  check_read_timeout ();	/* set by the read builtin */
   QUIT;
 }
 
@@ -981,7 +1006,7 @@ run_exit_trap ()
 	}
       else if (code == ERREXIT)
 	retval = last_command_exit_value;
-      else if (code == EXITPROG)
+      else if (code == EXITPROG || code == EXITBLTIN)
 	retval = last_command_exit_value;
       else if (function_code != 0)
         retval = return_catch_value;
@@ -1365,6 +1390,77 @@ restore_original_signals ()
   reset_or_restore_signal_handlers (restore_signal);
 }
 
+/* Change the flags associated with signal SIG without changing the trap
+   string. The string is TRAP_LIST[SIG] if we need it. */
+static void
+reinit_trap (sig)
+     int sig;
+{
+  sigmodes[sig] |= SIG_TRAPPED;
+  if (trap_list[sig] == (char *)IGNORE_SIG)
+    sigmodes[sig] |= SIG_IGNORED;
+  else
+    sigmodes[sig] &= ~SIG_IGNORED;
+  if (sigmodes[sig] & SIG_INPROGRESS)
+    sigmodes[sig] |= SIG_CHANGED;
+}
+
+/* Undo the effects of reset_signal_handlers(), which unsets the traps but
+   leaves the trap strings in place. This understands how reset_signal_handlers
+   works. */
+void
+restore_traps ()
+{
+  char *trapstr;
+  int i;
+
+  /* Take care of the exit trap first. If TRAP_LIST[0] is non-null, the trap
+     has been set. */
+  trapstr = trap_list[EXIT_TRAP];
+  if (trapstr)
+    reinit_trap (EXIT_TRAP);
+
+  /* Then DEBUG, RETURN, and ERROR. TRAP_LIST[N] == 0 if these signals are
+     not trapped. This knows what reset_signal_handlers does for these traps */
+  trapstr = trap_list[DEBUG_TRAP];
+  if (trapstr && function_trace_mode == 0)
+    reinit_trap (DEBUG_TRAP);
+  trapstr = trap_list[RETURN_TRAP];
+  if (trapstr && function_trace_mode == 0)
+    reinit_trap (RETURN_TRAP);
+  trapstr = trap_list[ERROR_TRAP];
+  if (trapstr && error_trace_mode == 0)
+    reinit_trap (ERROR_TRAP);
+
+  /* And finally all the `real' signals. reset_signal_handlers just changes the
+     signal handler for these signals, leaving the trap value in place. We
+     intuit what to do based on that value. We assume that signals marked as
+     SIG_SPECIAL are reinitialized by initialize_signals (), so we don't
+     change the signal handler unless the signal is supposed to be ignored. */
+  for (i = 1; i < NSIG; i++)
+    {
+      trapstr = trap_list[i];
+      if (sigmodes[i] & SIG_SPECIAL)
+	{
+	  if (trapstr && trapstr != (char *)DEFAULT_SIG)
+	    reinit_trap (i);
+	  if (trapstr == (char *)IGNORE_SIG && (sigmodes[i] & SIG_NO_TRAP) == 0)
+	    set_signal_handler (i, SIG_IGN);
+	}
+      else if (trapstr == (char *)IGNORE_SIG)
+	{
+	  reinit_trap (i);
+	  if ((sigmodes[i] & SIG_NO_TRAP) == 0)
+	    set_signal_handler (i, SIG_IGN);
+	}
+      else if (trapstr != (char *)DEFAULT_SIG)
+        /* set_signal duplicates the string argument before freeing it. */
+	set_signal (i, trapstr);
+
+      pending_traps[i] = 0;	/* XXX */
+    }
+}
+
 /* If a trap handler exists for signal SIG, then call it; otherwise just
    return failure.  Returns 1 if it called the trap handler. */
 int
@@ -1455,7 +1551,7 @@ signal_in_progress (sig)
   return (sigmodes[sig] & SIG_INPROGRESS);
 }
 
-#if 0 /* TAG: bash-5.2 */
+#if 0 /* unused */
 int
 block_trapped_signals (maskp, omaskp)
      sigset_t *maskp;
